package org.cobbzilla.wizard.cache.redis;

import lombok.Getter;
import lombok.NoArgsConstructor;
import lombok.NonNull;
import lombok.Setter;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import redis.clients.jedis.Jedis;
import redis.clients.jedis.params.SetParams;

import javax.annotation.Nullable;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import static org.cobbzilla.util.daemon.ZillaRuntime.*;
import static org.cobbzilla.util.json.JsonUtil.fromJsonOrDie;
import static org.cobbzilla.util.json.JsonUtil.toJsonOrDie;
import static org.cobbzilla.util.security.CryptoUtil.string_decrypt;
import static org.cobbzilla.util.security.CryptoUtil.string_encrypt;
import static org.cobbzilla.util.system.Sleep.sleep;

@Service @NoArgsConstructor @Slf4j
public class RedisService {

    public static final byte MAX_RETRIES = 5;

    public static final String NX = "NX";
    public static final String XX = "XX";

    public static final String EX = "EX";
    public static final String PX = "PX";

    public static final String ALL_KEYS = "*";

    @Autowired @Getter @Setter private HasRedisConfiguration configuration;

    @Getter @Setter private String key;
    protected boolean hasKey() { return !empty(getKey()); }

    private final AtomicReference<Jedis> redis = new AtomicReference<>();
    private Jedis newJedis() {
        return new Jedis(configuration.getRedis().getHost(), configuration.getRedis().getPort());
    }

    @Getter @Setter private String prefix = null;

    public RedisService(HasRedisConfiguration configuration) {
        this(configuration, configuration.getRedis().getPrefix(), configuration.getRedis().getKey());
    }

    public RedisService(HasRedisConfiguration configuration, String prefix, String key) {
        this(configuration.getRedis(), prefix, key);
    }

    public RedisService(final RedisConfiguration configuration, String prefix, String key) {
        this.configuration = () -> configuration;
        this.prefix = prefix;
        this.key = key;
    }

    private Map<String, RedisService> prefixServiceCache = new ConcurrentHashMap<>();

    public RedisService prefixNamespace(String prefix) {
        return prefixNamespace(prefix, configuration.getRedis().getKey());
    }

    public RedisService prefixNamespace(String prefix, String key) {
        RedisService r = prefixServiceCache.get(prefix);
        if (r == null) {
            String basePrefix = (this.prefix != null) ? this.prefix : configuration.getRedis().getPrefix();
            basePrefix = empty(basePrefix) ? "" : basePrefix + ".";
            r = new RedisService(configuration, basePrefix + prefix, key);
            prefixServiceCache.put(prefix, r);
        }
        return r;
    }

    public void reconnect() {
        if (log.isDebugEnabled()) log.debug("marking redis for reconnection...");
        synchronized (redis) {
            if (redis.get() != null) {
                try { redis.get().disconnect(); } catch (Exception e) {
                    log.warn("error disconnecting from redis before reconnecting: "+e);
                }
            }
            redis.set(null);
        }
    }

    private Jedis getRedis() {
        synchronized (redis) {
            if (redis.get() == null) {
                if (log.isDebugEnabled()) log.debug("connecting to redis...");
                redis.set(newJedis());
            }
        }
        return redis.get();
    }

    public <V> RedisMap<V> map(String prefix) { return map(prefix, null); }
    public <V> RedisMap<V> map(String prefix, Long duration) { return new RedisMap<>(prefix, duration, this); }

    public boolean exists(@NonNull final String key) {
        return retry(() -> getRedis().exists(prefix(key)), "exists");
    }

    public boolean anyExists(@NonNull final Collection<String> keys) {
        for (String k : keys) if (exists(k)) return true;
        return false;
    }

    public boolean allExist(@NonNull final Collection<String> keys) {
        final var prefixedKeysArray = keys.stream().map(k -> prefix(k)).toArray(String[]::new);
        return keys.size() == retry(() -> getRedis().exists(prefixedKeysArray), "exists");
    }

    public <T> T getObject(String key, Class<T> clazz) {
        final String json = get(key);
        return empty(json) ? null : fromJsonOrDie(json, clazz);
    }

    @Nullable public String get(@NonNull final String key) {
        return decrypt(retry(() -> getRedis().get(prefix(key)), "get"));
    }
    @Nullable public String get_withPrefix(@NonNull final String prefixedKey) {
        return decrypt(retry(() -> getRedis().get(prefixedKey), "get"));
    }
    @Nullable public String get_plaintext(@NonNull final String key) {
        return retry(() -> getRedis().get(prefix(key)), "get");
    }


    public void set(@NonNull final String key, @NonNull final String value) {
        retry(() -> getRedis().set(prefix(key), encrypt(value)), "set");
    }
    public <T> void setObject(@NonNull final String key, @NonNull final T thing) { set(key, toJsonOrDie(thing)); }

    public void set(@NonNull final String key, @NonNull final String value, @NonNull final String nxxx,
                    @NonNull final String expx, final long time) {
        retry(() -> getRedis().set(prefix(key), encrypt(value), buildSetParams(nxxx, expx, time)), "set");
    }
    public void set(@NonNull final String key, @NonNull final String value, @NonNull final String expx,
                    final long time) {
        set(key, value, XX, expx, time);
        set(key, value, NX, expx, time);
    }
    public void setAll(@NonNull final Collection<String> keys, @NonNull final String value, @NonNull final String expx,
                       final long time) {
        for (String k : keys) set(k, value, expx, time);
    }

    public void set_plaintext(@NonNull final String key, @NonNull final String value) {
        retry(() -> getRedis().set(prefix(key), value), "set_plaintext");
    }

    public void set_plaintext(@NonNull final String key, @NonNull final String value, @NonNull final String nxxx,
                              @NonNull final String expx, final long time) {
        retry(() -> getRedis().set(prefix(key), value, buildSetParams(nxxx, expx, time)), "set_plaintext");
    }

    public void set_plaintext(@NonNull final String key, @NonNull final String value, @NonNull final String expx,
                              final long time) {
        set_plaintext(key, value, XX, expx, time);
        set_plaintext(key, value, NX, expx, time);
    }

    public Long touch(@NonNull final String key) { return retry(() -> getRedis().touch(prefix(key)), "touch"); }

    public Long expire(@NonNull final String key, final int ttlSeconds) {
        return retry(() -> getRedis().expire(prefix(key), ttlSeconds), "expire");
    }

    public Long pexpire(@NonNull final String key, final long ttlMillis) {
        return retry(() -> getRedis().pexpire(prefix(key), ttlMillis), "pexpire");
    }

    public List<String> list(@NonNull final String key) { return lrange(key, 0, -1); }
    public List<String> lrange(@NonNull final String key, final int start, final int end) {
        final var list = retry(() -> getRedis().lrange(prefix(key), start, end), "lrange");
        return hasKey() ? list.stream().map(this::decrypt).collect(Collectors.toList()) : list;
    }

    public Long llen(@NonNull final String key) {
        return retry(() -> getRedis().llen(prefix(key)), "llen");
    }
    public void lpush(@NonNull final String key, @NonNull final String value) {
        retry(() -> getRedis().lpush(prefix(key), encrypt(value)), "lpush");
    }
    public String lpop(@NonNull final String key) {
        return decrypt(retry(() -> getRedis().lpop(prefix(key)), "lpop"));
    }
    public void rpush(@NonNull final String key, @NonNull final String value) {
        retry(() -> getRedis().rpush(prefix(key), encrypt(value)), "rpush");
    }
    public String rpop(@NonNull final String key) {
        return decrypt(retry(() -> getRedis().rpop(prefix(key)), "rpop"));
    }

    public void hset(@NonNull final String key, @NonNull final String field, @NonNull final String value) {
        retry(() -> getRedis().hset(prefix(key), encrypt(field), encrypt(value)), "hget");
    }
    public String hget(@NonNull final String key, @NonNull final String field) {
        if (empty(field)) return die("__hget(" + key + "/): field was empty");
        return decrypt(retry(() -> getRedis().hget(prefix(key), encrypt(field)), "hget"));
    }
    public Map<String, String> hgetall(@NonNull final String key) {
        final var map = retry(() -> getRedis().hgetAll(prefix(key)), "hgetall");
        if (!hasKey()) return map;

        final var decrypted = new HashMap<String, String>(map.size());
        for (final var entry : map.entrySet()) {
            decrypted.put(decrypt(entry.getKey()), decrypt(entry.getValue()));
        }
        return decrypted;
    }
    public Long hdel(@NonNull final String key, @NonNull final String field) {
        return retry(() -> getRedis().hdel(prefix(key), encrypt(field)), "hdel");
    }
    public Set<String> hkeys(@NonNull final String key) {
        return retry(() -> getRedis().hkeys(prefix(key)), "hget");
    }
    public Long hlen(@NonNull final String key) {
        return retry(() -> getRedis().hlen(prefix(key)), "hlen");
    }

    public Long del(@NonNull final String key) {
        return retry(() -> getRedis().del(prefix(key)), "del");
    }
    public Long del_withPrefix(@NonNull final String prefixedKey) {
        return retry(() -> getRedis().del(prefixedKey), "del");
    }
    public Long del_matching(@NonNull final String keyMatch) {
        var count = 0L;
        for (final var key : keys(keyMatch)) {
            count += del_withPrefix(key);
        }
        return count;
    }

    public Long sadd(@NonNull final String key, final String value) { return sadd(key, new String[]{value}); }
    public Long sadd(@NonNull final String key, final String[] values) {
        return retry(() -> getRedis().sadd(prefix(key), encrypt(values)), "sadd");
    }
    public Long sadd_plaintext(@NonNull final String key, final String value) {
        return sadd_plaintext(key, new String[]{value});
    }
    public Long sadd_plaintext(@NonNull final String key, final String[] values) {
        return retry(() -> getRedis().sadd(prefix(key), values), "sadd");
    }
    public Long srem(@NonNull final String key, final String value) { return srem(key, new String[]{value}); }
    public Long srem(@NonNull final String key, final String[] values) {
        return retry(() -> getRedis().srem(prefix(key), encrypt(values)), "srem");
    }
    public Set<String> smembers(@NonNull final String key) {
        return retry(() -> decrypt(getRedis().smembers(prefix(key))), "smembers");
    }
    public boolean sismember(@NonNull final String key, @NonNull final String value) {
        return retry(() -> getRedis().sismember(prefix(key), encrypt(value)), "sismember");
    }
    public List<String> srandmembers(@NonNull final String key, final int count) {
        return retry(() -> decrypt(getRedis().srandmember(prefix(key), count)), "srandmember");
    }
    public String srandmember(@NonNull final String key) {
        final var rand = srandmembers(key, 1);
        return empty(rand) ? null : rand.get(0);
    }
    @Nullable public Set<String> spop(@NonNull final String key, final long count) {
        return retry(() -> decrypt(getRedis().spop(prefix(key), count)), "spop");
    }
    @Nullable public String spop(@NonNull final String key) {
        final var member = spop(key, 1);
        return empty(member) ? null : member.iterator().next();
    }
    public long scard(@NonNull final String key) {
        return retry(() -> getRedis().scard(prefix(key)), "scard");
    }
    public Set<String> sunion(@NonNull final Collection<String> keys) {
        return decrypt(sunion_plaintext(keys));
    }
    public Set<String> sunion_plaintext(@NonNull final Collection<String> keys) {
        String[] prefixedKeys = keys.stream().map(this::prefix).toArray(String[]::new);
        return retry(() -> getRedis().sunion(prefixedKeys), "sunion");
    }
    public Long sunionstore((@NonNull final String destKey, (@NonNull final Collection<String> keys) {
        String[] prefixedKeys = keys.stream().map(this::prefix).toArray(String[]::new);
        return retry(() -> getRedis().sunionstore(prefix(destKey), prefixedKeys), "sunionstore");
    }

    @Nullable public Long counterValue(@NonNull final String key) {
        final var value = get_plaintext(key);
        return value == null ? null : Long.parseLong(value);
    }
    public Long incr(@NonNull final String key) {
        return retry(() -> getRedis().incrBy(prefix(key), 1), "incrBy");
    }
    public Long incrBy(@NonNull final String key, final long value) {
        return retry(() -> getRedis().incrBy(prefix(key), value), "incrBy");
    }
    public Long decr(@NonNull final String key) {
        return retry(() -> getRedis().decrBy(prefix(key), 1), "decrBy");
    }
    public Long decrBy(@NonNull final String key, final long value) {
        return retry(() -> getRedis().decrBy(prefix(key), value), "decrBy");
    }

    public Collection<String> keys(@NonNull final String key) {
        return retry(() -> getRedis().keys(prefix(key)), "keys");
    }

    public String rename(@NonNull final String key, @NonNull final String newKey) {
        retry(() -> getRedis().rename(prefix(key), newKey), "rename");
    }

    public static final String LOCK_SUFFIX = "._lock";

    public boolean confirmLock(String key, String lock, long lockTimeout) {
        key = key + LOCK_SUFFIX;
        final String lockVal = get(key);
        if (lockVal != null && lockVal.equals(lock)) {
            pexpire(key, lockTimeout);
            return true;
        }
        return false;
    }

    public String lock(String key, long lockTimeout, long deadlockTimeout) {
        if (log.isDebugEnabled()) log.debug("lock("+key+") starting");
        key = key + LOCK_SUFFIX;
        final String uuid = UUID.randomUUID().toString();
        String lockVal = get(key);
        final long start = now();
        while ((lockVal == null || !lockVal.equals(uuid)) && (now() - start < lockTimeout)) {
            set(key, uuid, NX, EX, deadlockTimeout/1000);
            if (log.isDebugEnabled()) log.debug("lock("+key+") locked with uuid="+uuid);
            lockVal = get(key);
            if (log.isDebugEnabled()) log.debug("lock("+key+") after locking with uuid="+uuid+", lockVal="+lockVal);
        }
        if (lockVal == null || !lockVal.equals(uuid)) {
            return die("lock: timeout locking "+key);
        }
        if (log.isDebugEnabled()) log.debug("lock: LOCKED "+key+" = "+lockVal);
        return uuid;
    }

    public void unlock(String key, String lock) {
        key = key + LOCK_SUFFIX;
        final String lockVal = get(key);
        if (lockVal == null || !lockVal.equals(lock)) {
            log.warn("unlock: already unlocked! "+key);
        } else {
            del(key);
            log.info("unlock: UNLOCKED "+key);
        }
    }

    public String loadScript(final String script) {
        return retry(() -> getRedis().scriptLoad(script), "loadScript");
    }


    public Object eval(final String scriptSHA, final List<String> keys, final List<String> args) {
        return retry(() -> getRedis().evalsha(scriptSHA, prefix(keys), args), "eval");
    }
    public String prefix(String key) { return empty(prefix) ? key : prefix + "." + key; }

    public List<String> prefix(Collection<String> keys) {
        return keys.parallelStream().map(this::prefix).collect(Collectors.toList());
    }
    // override these for full control

    protected String encrypt(String data) {
        if (!hasKey()) return data;
        return string_encrypt(data, getKey());
    }

    protected String[] encrypt(final String[] data) {
        if (!hasKey() || empty(data)) return data;
        final String[] encrypted = new String[data.length];
        for (int i=0; i<data.length; i++) {
            encrypted[i] = string_encrypt(data[i], getKey());
        }
        return encrypted;
    }

    protected String decrypt(String data) {
        if (!hasKey()) return data;
        if (data == null) return null;
        return string_decrypt(data, getKey());
    }
    protected <T extends Collection<String>> T decrypt(T data) {
        if (!hasKey() || empty(data)) return data;
        final T decrypted = (T) new ArrayList<String>();
        for (String value : data) {
            decrypted.add(string_decrypt(value, getKey()));
        }
        return decrypted;
    }

    protected String[] decrypt(String[] data) {
        if (!hasKey() || empty(data)) return data;
        final String[] decrypted = new String[data.length];
        for (int i=0; i<data.length; i++) {
            decrypted[i] = string_decrypt(data[i], getKey());
        }
        return decrypted;
    }

    @Nullable private <T> T retry(@NonNull final Supplier<T> function, @NonNull final String logSuffix) {
        for (var attempt = 1; attempt <= MAX_RETRIES; attempt++) {
            try {
                synchronized (redis) {
                    return function.get();
                }
            } catch (RuntimeException e) {
                if (attempt >= MAX_RETRIES) throw e; // here's the final exit from the loop (no need to reconnect)
                reconnect();
                sleep(attempt * 10, "retrying RedisService " + logSuffix);
            }
        }

        // Should not happen, but just in case:
        throw new RuntimeException("retry: Something went wrong while working with redis");
    }

    private SetParams buildSetParams(String nxxx, String expx, long time) {
        SetParams setParams = new SetParams();
        switch (nxxx) {
            case NX: setParams.nx(); break;
            case XX: setParams.xx(); break;
        }
        switch (expx) {
            case EX: setParams.ex((int) time); break;
            case PX: setParams.px(time); break;
        }
        return setParams;
    }

    public void flush() { del_matching(ALL_KEYS); }

}
